#
# Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
# All rights reserved.
# This component and the accompanying materials are made available
# under the terms of "Eclipse Public License v1.0"
# which accompanies this distribution, and is available
# at the URL "http://www.eclipse.org/legal/epl-v10.html".
#
# Initial Contributors:
# Nokia Corporation - initial contribution.
#
# Contributors:
#
# Description:
# All blocks commands are here
#

import logging
import os
import tarfile
import collections
import urlparse

from Blocks.debfile import DebFile, DebError

from UserConfiguration.UserConfiguration import BadWorkspace, UserConfiguration, WorkspaceConfigError
from sourceslist import SourcesList, RepoError
from workspacediff import WorkspaceDiff
from generalexceptions import ExternalProcessError, InvalidFileInput, InvalidUserInput, GenericError
import utils
import uitools
import localbundle
import cmdlineutils
from cmdlinecommand import CmdLineCommand, CommandOptionParser
import comps

class BlocksCommand(CmdLineCommand):
    ''' Blocks commands '''

    ENV_WSID = "BLOCKS_WSID"
    WS_NOT_DEFINED_ERROR = ("Workspace not defined.\n"
                            "Please go to workspace directory, set environment variable '%s', use options or "
                            "set default workspace with workspace-select." % ENV_WSID)
    NO_REPOS_DEFINED = "No repositories defined for the current workspace."

    def checkWorkspace(self, idnum=None):
        idnum = idnum if idnum is not None else self.workspaceId
        if idnum is None:
            raise BadWorkspace(self.WS_NOT_DEFINED_ERROR)
        try:
            idnum = int(idnum)
        except ValueError:
            raise BadWorkspace("Workspace id must be integer")
        self.config.checkWorkspaceExists(idnum)

    def __init__(self, options):
        CmdLineCommand.__init__(self, options)
        # If these are called uninitialized return error string
        # When workspace is not really needed handle it in the command
        self.dpkg = utils.getErrorFunction(self.WS_NOT_DEFINED_ERROR)
        self.apt = utils.getErrorFunction(self.WS_NOT_DEFINED_ERROR)

        self.config = UserConfiguration()
        self.workspaceId = None

        uitools.force = True if self.options.force or self.options.verbose <= -2 else False

        self._setCurrentWorkspace()

    def _setCurrentWorkspace(self):
        wsid = os.environ.get(BlocksCommand.ENV_WSID)
        if self.options.wsid != None:
            wsid = self.options.wsid
        elif self.options.wsname != None:
            wsid = self.config.getWorkspaceIdByName(self.options.wsname)
            if wsid == None:
                raise BadWorkspace("Workspace with name '%s' does not exist" % self.options.wsname)
        elif wsid:
            if not wsid.isdigit() or not self.config.workspaceExists(int(wsid)):
                logging.warning("Invalid workspace id '%s' in environment variable '%s'", wsid, self.ENV_WSID)
                wsid = None
        elif self.config.default_workspace > 0:
            wsid = self.config.default_workspace
        else:
            ws = self.config.getWorkspaces()
            if ws:
                # If current working dir is on substed drive get the real path
                cwd = os.path.normcase(utils.getRealPath(os.getcwd())) + os.sep
                for idnum in ws:
                    path = os.path.normcase(self.config.getWorkspacePath(idnum))
                    path = utils.addPathSep(path)
                    if cwd.startswith(path):
                        wsid = idnum
                        logging.info("Workspace '%s' with id %s selected because of current directory.", self.config.getWorkspaceNameById(wsid), wsid)
                        break
        if wsid != None:
            self.setWorkspace(int(wsid))

    def setWorkspace(self, idnum=None):
        '''
        Set workspace to be used
        If workspace is not needed call without id
        '''
        if idnum is None:
            try:
                # Check first if there is already a valid workspace
                self.checkWorkspace()
            except BadWorkspace:
                self.dpkg = cmdlineutils.Dpkg()
                self.apt = cmdlineutils.Apt()
        else:
            self.workspaceId = idnum
            wsPath = self.config.getWorkspaceById(idnum).path
            wsMetaPath = self.config.getWorkspaceMetadataPath(self.workspaceId)
            self.dpkg = cmdlineutils.Dpkg(wsPath, wsMetaPath, self.options)
            self.apt = cmdlineutils.Apt(wsPath, wsMetaPath, self.options)

    def cmd_workspace_add(self, path):
        ''' Add new workspace '''
        path = os.path.abspath(path)
        if not os.path.isdir(path):
            raise BadWorkspace("Directory '%s' does not exist" % path)
        realpath = utils.getRealPath(path)
        if realpath != path:
            logging.info("Path '%s' on substed drive. Adding workspace to real path '%s'.", path, realpath)
            path = realpath

        name = self.cmd_options.name
        if not name:
            name = self.getUniqueWorkspaceName()

        self.config.createWorkspace(name, path)
        print "Workspace id:", self.config.getWorkspaceIdByName(name)

    cmd_workspace_add._optionparser = CommandOptionParser()
    cmd_workspace_add._optionparser.add_option("-n", "--name", help="Workspace name [default: Generated]")

    def getUniqueWorkspaceName(self):
        wsnames = [self.config.getWorkspaceNameById(wsid) for wsid in self.config.getWorkspaces()]
        return utils.uniqueName("workspace", wsnames)

    def getWorkspaceId(self, id_or_name):
        if id_or_name.isdigit():
            idnum = id_or_name
        else:
            idnum = self.config.getWorkspaceIdByName(id_or_name)
            if idnum is None:
                raise BadWorkspace("Workspace with name '%s' does not exist" % id_or_name)
        return int(idnum)

    def cmd_workspace_select(self, id_or_name):
        ''' Select default workspace (deselect with id 0) '''
        self.config.setDefaultWorkspace(self.getWorkspaceId(id_or_name))

    def cmd_workspace_list(self):
        ''' List workspaces and show the current one '''
        ws = self.config.getWorkspaces()
        if ws:
            for idnum in ws:
                print "%s%s%s\n  name: %s\n  path: %s\n" % (idnum,
                                                            "*" if idnum == self.workspaceId else "",
                                                            " (default)" if idnum == self.config.default_workspace else "",
                                                            self.config.getWorkspaceById(idnum).name,
                                                            self.config.getWorkspaceById(idnum).path)
        else:
            print "No workspaces defined."

    def cmd_workspace_remove(self, id_or_name):
        ''' Remove workspace (removes only metadata) '''
        idnum = self.getWorkspaceId(id_or_name)
        self.checkWorkspace(idnum)

        text = """Removing workspace (only metadata).
Workspace info:
  Id: %d
  Name: %s
  Path: %s
"""
        workspace = self.config.getWorkspaceById(idnum)
        if uitools.askConfirmation(text % (idnum, workspace.name, workspace.path)):
            self.config.removeWorkspace(int(idnum))

    def cmd_advanced_metadata_check(self): # pragma: no cover
        ''' Check inconsistencies in metadata '''
        oL = self.config.getOrphanedListEntries()
        if oL:
            print "Found workspace entries without matching metadata directories:"
            for o in oL:
                print "\tID: %s\tMissing directory: %s" % (o, self.config.getWorkspaceMetadataPath(o))
            print

        oM = self.config.getOrphanedMetadataDirs()
        if oM:
            print "Found unlisted metadata directories:"
            for o in oM:
                print "\t%s" % (o)
            print

        oW = self.config.getOrphanedWorkspaces()
        if oW:
            print "Found metadata for non-existent workspaces:"
            for o in oW:
                print "\tID: %s\tName: '%s'\tPath: %s" % (o, self.config.getWorkspaceNameById(o), self.config.getWorkspaceById(o).path)

    def aptUpdate(self):
        self.checkWorkspace()
        try:
            slist = SourcesList(self.getSourcesListPath())
            if not slist.repos:
                raise GenericError(self.NO_REPOS_DEFINED)
            logging.debug("\n" + self.apt("update", quiet=True))
            # get comps.xmls
            utils.createDir(self.getCompsPath())
            logging.debug("Starting to retrieve comps")
            for name, uri in slist.repos:
                uri = utils.addSuffix(uri, "/")
                uri = urlparse.urljoin(uri, "comps.xml")
                localpath = self.getRepoCompsPath(name)
                logging.debug("Retrieving '%s' to '%s'", uri, localpath)
                try:
                    utils.urlretrieve(uri, localpath)
                except IOError, ex:
                    if ex.errno == 404:
                        logging.debug(str(ex))
                    else:
                        logging.warning(str(ex))
        except ExternalProcessError, ex:
            logging.warning("Bundle list update failed: %s\nRun apt-get update for more information.", ex)

    def getCompsPath(self):
        wsMetaPath = self.config.getWorkspaceMetadataPath(self.workspaceId)
        return os.path.join(wsMetaPath, "comps")

    def getRepoCompsPath(self, reponame):
        compsDir = self.getCompsPath()
        return os.path.join(compsDir, reponame + ".xml")

    def getCurrentCompsPaths(self):
        slist = SourcesList(self.getSourcesListPath())
        return [self.getRepoCompsPath(name) for name, _ in slist.repos]

    @staticmethod
    def _separateDebs(files):
        ''' Separate names which are deb files '''
        bundles = []
        debs = []
        for name in files:
            if name.endswith(".deb"):
                debs.append(name)
            else:
                bundles.append(name)
        return bundles, debs

    def cmd_bundle_install(self, bundle1, *bundle):
        ''' Install/update bundle(s) from repo/file/http/ftp '''
        self.checkWorkspace()
        bundles, debs = self._separateDebs(utils.listify(bundle1, bundle))
        if bundles:
            self.aptUpdate()
            self.apt("install %s" % " ".join(bundles))
            # NOTE: If apt raises an exception we won't clean up the cache
            self.cleanCache()
        if debs:
            lb = localbundle.LocalBundle(self.dpkg)
            try:
                lb.install(debs)
            except ValueError, ex:
                raise IOError("None of the deb files found. %s." % ex)

    def getGroupComp(self):
        self.aptUpdate()
        comp = comps.Comps()
        for path in self.getCurrentCompsPaths():
            try:
                comp.add(path)
            except SyntaxError, ex:
                raise InvalidFileInput("Error in comps.xml: %s" % ex)
            except IOError, ex: # comps.xml not available on repo
                logging.debug("Cannot get comps.xml: %s", ex)
        comp.resolveInstalledGroups(self.getInstalledPackages())
        return comp

    def groupProcess(self, groups, processor):
        comp = self.getGroupComp()
        for name in groups:
            group = comp.getGroup(name)
            if group:
                processor(group)
            else:
                raise InvalidUserInput("Group '%s' not found" % name)

    def cmd_group_list(self):
        ''' List available/installed bundle groups '''
        comp = self.getGroupComp()
        installedGroups = ["  %s" % g.name for g in comp.groups.Installed]
        availableGroups = ["  %s" % g.name for g in comp.groups.Available]
        result = []
        if installedGroups:
            result += ["Installed Groups:"] + sorted(installedGroups)
        if availableGroups:
            if installedGroups:
                result += [""]
            result += ["Available Groups:"] + sorted(availableGroups)
        if not result:
            result = ["No groups available on current repositories"]
        print "\n".join(result)

    @staticmethod
    def _askGroupTypeInstall(group, typeText):
        packages = group.packages[typeText]
        packageCount = len(packages)
        if packageCount > 0:
            result = "s"
            while result == "s":
                result = uitools.ask("""Group contains %d %s bundles.
Do you want to install those (answer 's' to show bundles)""" % (packageCount, typeText), ["y", "n", "s"])
                if result == "s":
                    print "\n%s\n" % "\n".join(packages)
            return result == "y"
        else:
            return False

    def cmd_group_install(self, group1, *group):
        ''' Install bundle group(s) '''
        def processor(group):
            if not group.installed:
                installPackages = group.mandatoryPackages
                defaultValues = self.cmd_group_install._optionparser.get_default_values()
                def processGroupType(gtype):
                    install = getattr(self.cmd_options, gtype) == "y"
                    ask = getattr(defaultValues, gtype) == "n" and not install
                    if ask:
                        install = self._askGroupTypeInstall(group, gtype)
                    if install:
                        return list(group.packages[gtype])
                    else:
                        return []

                installPackages += processGroupType("default")
                installPackages += processGroupType("optional")

                self.apt("install " + utils.argsToStr(installPackages))
                self.cleanCache()
            else:
                print "Group '%s' is already installed." % group.name

        self.groupProcess(utils.listify(group1, group), processor)

    cmd_group_install._optionparser = CommandOptionParser()
    cmd_group_install._optionparser.add_option("-d", "--default", choices=["y", "n"], metavar="y/n",
                                               help="Install default bundles [default: %default]")
    cmd_group_install._optionparser.add_option("-o", "--optional", choices=["y", "n"], metavar="y/n",
                                               help="Install optional bundles [default: %default]")
    cmd_group_install._optionparser.set_defaults(default="y", optional="n")

    def cmd_group_remove(self, group1, *group):
        ''' Remove bundle group(s) '''
        def processor(group):
            if group.installed:
                self.apt("remove " + utils.argsToStr(group.allPackages))
            else:
                print "Group '%s' is not installed." % group.name

        self.groupProcess(utils.listify(group1, group), processor)

    # TODO: Add option to show package installation status?
    def cmd_group_info(self, group1, *group):
        ''' Get information about group(s) '''
        def processor(group):
            formatString = "Group: %s\n  Description: %s\n  Installed: %s"
            print formatString % (group.name,
                                  group.description,
                                  "yes" if group.installed else "no")
            for packageType in ((group.mandatoryPackages, "Mandatory Bundles"),
                                (group.defaultPackages, "Default Bundles"),
                                (group.optionalPackages, "Optional Bundles")):
                if packageType[0]:
                    print "\n  %s:" % packageType[1]
                    print "\n".join("    %s" % p for p in packageType[0])
            print

        self.groupProcess(utils.listify(group1, group), processor)

    # TODO: Refactor this with workspacediff
    class Output(object):

        def __init__(self, differ):
            self.differ = differ
            self.bundlestatus = collections.defaultdict(set)

        def __call__(self, status, bundle, name):
            self.bundlestatus[bundle].add(status)
            if ("M" in self.bundlestatus[bundle] or
                "R" in self.bundlestatus[bundle]):
                return True

    def getModifiedPackages(self, packages=None):
        wsPath = self.config.getWorkspaceById(self.workspaceId).path
        wsMetaPath = self.config.getWorkspaceMetadataPath(self.workspaceId)
        wsdiff = WorkspaceDiff(wsPath, wsMetaPath, self.Output)
        wsdiff.start(packages, bundlediff=True)
        return [k for k, v in wsdiff.output.bundlestatus.iteritems() if "M" in v or "R" in v]

    def cmd_workspace_restore(self, *bundle):
        ''' Reinstall all bundles or just the specified ones '''
        self.aptUpdate()
        if self.options.force:
            if bundle:
                packages = bundle
            else:
                packages = self.getInstalledPackages()
        else:
            packages = self.getModifiedPackages(bundle)
        if packages:
            self.apt("--reinstall install " + utils.argsToStr(packages))
            self.cleanCache()
        else:
            print "Modified bundles not found."

    def cmd_update(self):
        ''' Update all bundles and dependencies '''
        self.aptUpdate()
        self.apt("dist-upgrade")
        self.cleanCache()

    def cmd_bundle_remove(self, bundle1, *bundle):
        ''' Remove bundle(s) '''
        self.apt("remove %s" % utils.argsToStr(bundle1, bundle))

    def cmd_bundle_info(self, bundle1, *bundle):
        ''' Show information about bundles and .deb files '''
        bundles, debs = self._separateDebs(utils.listify(bundle1, bundle))
        for name in debs:
            try:
                deb = DebFile(name)
            except DebError, ex:
                logging.warning("Could not open '%s': %s", name, ex)
            else:
                metadata = deb.metadata
                names = [
                    "Package", "Version", "Architecture", "Installed-Size", "Depends",
                    "Replaces", "Provides", "Conflicts", "Priority", "Description"]
                names += sorted(name for name in metadata if name.startswith("X-"))
                for name in names:
                    print "%s: %s" % (name, metadata.get(name, "n/a"))
            print
        if bundles:
            self.aptUpdate()
            for bundle in bundles:
                # Apt seems to give us extra blank line after metadata -> remove
                print self.apt("show %s" % bundle, "apt-cache", True).strip()
                print

    def cmd_search(self, pattern):
        ''' Search for bundles from repositories '''
        self.aptUpdate()
        self.apt("search " + pattern, "apt-cache")

    def cmd_repo_add(self, uri):
        ''' Add repository to current workspace '''
        self.checkWorkspace()
        slist = SourcesList(self.getSourcesListPath())
        uri = uri.replace("\\", "/")
        name = self.cmd_options.name
        if not name:
            name = utils.uniqueName("repo", [repo.name for repo in slist.repos])
        try:
            slist.add(name, uri)
        except RepoError, ex:
            raise InvalidUserInput(str(ex))
        uri = utils.addSuffix(uri, "/")
        try:
            utils.openUrl(urlparse.urljoin(uri, "Packages")).close()
        except IOError, ex:
            problem = "Repository you are about to add has a problem:\nCannot get the index file: %s" % ex
            if not uitools.askConfirmation(problem):
                slist.removeByName(name)
        # Test updating
        self.aptUpdate()

    cmd_repo_add._optionparser = CommandOptionParser()
    cmd_repo_add._optionparser.add_option("-n", "--name", help="Repository name [default: Generated]")

##    def aptUpdateTest(self):
##        try:
##            output = self.apt("update", quiet=True)
##            return (not any(line.startswith("Err") for line in output.splitlines()), output)
##        except ExternalProcessError, ex:
##            return (False, ex.output)

    def cmd_repo_remove(self, id_or_name):
        ''' Remove repository from current workspace '''
        self.checkWorkspace()
        slist = SourcesList(self.getSourcesListPath())
        try:
            if id_or_name.isdigit():
                idnum = id_or_name
            else:
                idnum = slist.getRepoIdByName(id_or_name)
            repo = slist.getRepo(idnum)
        except RepoError, ex:
            raise InvalidUserInput(str(ex))
        # Remove possibly downloaded comps file
        try:
            os.remove(self.getRepoCompsPath(repo.name))
        except OSError:
            pass # comps.xml not found
        slist.remove(idnum)

    def cmd_repo_list(self):
        ''' List repositories in current workspace '''
        self.checkWorkspace()
        slist = SourcesList(self.getSourcesListPath())
        if slist.repos:
            print "\n\n".join("%d\n  Name: %s\n  URI: %s" % (i+1, name, uri)
                              for i, (name, uri) in enumerate(slist.repos))
        else:
            print self.NO_REPOS_DEFINED

    def cmd_advanced_bundle_depends(self, bundle1, *bundle):
        ''' List bundle dependencies '''
        self.apt("depends %s" % utils.argsToStr(bundle1, bundle), "apt-cache")

    def cmd_advanced_bundle_rdepends(self, bundle1, *bundle):
        ''' List bundle reverse dependencies '''
        self.apt("rdepends %s" % utils.argsToStr(bundle1, bundle), "apt-cache")

    def cmd_find_owner(self, path):
        ''' Find owner bundle of a file '''
        self.checkWorkspace()
        isPath = "\\" in path or "/" in path
        if isPath:
            path = os.path.normpath(os.path.abspath(path))
            wsPath = self.config.getWorkspacePath(self.workspaceId)
            if utils.pathInside(wsPath, path, False):
                path = utils.removeStart(path, utils.addPathSep(os.path.normpath(wsPath)))
            else:
                raise InvalidUserInput("Path not inside current workspace '%s'" % wsPath)
        path = path.replace("\\", "/")
        self.dpkg("-S " + path)

    def getInstalledPackages(self, arg=None):
        if arg == None:
            arg = []
        output = self.dpkg("--get-selections " + utils.argsToStr(arg), True)
        return [line.split()[0] for line in output.splitlines() if line.split()[1] == "install"]

    def cmd_bundle_list(self, *pattern):
        ''' List installed bundles (optional wildcard search) '''
        print "\n".join(self.getInstalledPackages(pattern))
        #self.dpkg("--list " + utils.argsToStr(arg)) # This gives more detailed stuff

    def cmd_bundle_list_files(self, bundle):
        ''' List files of a installed bundle or a .deb file '''
        if utils.isFile(bundle, ".deb"):
            self.setWorkspace() # no workspace needed
            self.dpkg("--contents " + '"' + bundle + '"')
        else:
            self.dpkg("--listfiles " + bundle)

    def cmd_advanced_apt_get(self, *arg):
        ''' Pass commands/options to apt-get '''
        self.setWorkspace() # no workspace needed
        self.apt(utils.argsToStr(arg))

    def cmd_advanced_apt_cache(self, *arg):
        ''' Pass commands/options to apt-cache '''
        self.setWorkspace() # no workspace needed
        self.apt(utils.argsToStr(arg), "apt-cache")

    def cmd_advanced_dpkg(self, *arg):
        ''' Pass commands/options to dpkg '''
        self.setWorkspace() # no workspace needed
        self.dpkg(utils.argsToStr(arg))

    def cmd_advanced_workspace_export(self, outputfile):
        '''Export workspace metadata to a file

        Workspace is selected by normal selection procedure.
        '''
        self.checkWorkspace()
        tar = tarfile.open(outputfile, "w:gz")
        tar.add(self.config.getWorkspaceMetadataPath(self.workspaceId), arcname="",
                exclude=lambda f: f.lower().endswith(".deb"))
        tar.close()

    def cmd_advanced_workspace_import(self, inputfile, path):
        ''' Import workspace metadata from a file '''
        try:
            tar = tarfile.open(inputfile, "r")
        except IOError:
            raise IOError("File '%s' does not exist." % inputfile)
        name = self.cmd_options.name
        if not name:
            name = self.getUniqueWorkspaceName()
        # This is for add workspace command
        self.cmd_options.name = name
        self.cmd_workspace_add(path)
        tar.extractall(self.config.getWorkspaceMetadataPath(self.config.getWorkspaceIdByName(name)))
        tar.close()

    cmd_advanced_workspace_import._optionparser = CommandOptionParser()
    cmd_advanced_workspace_import._optionparser.add_option("-n", "--name", help="Workspace name [default: Generated]")

    def cmd_advanced_workspace_property_list(self):
        ''' List workspace properties '''
        self.checkWorkspace()
        properties = self.config.getWorkspaceById(self.workspaceId).getPropertyValues()
        for k, v in properties.iteritems():
            print "%s = %s" % (k, v)

    def cmd_advanced_workspace_property_set(self, prop, value):
        ''' Set workspace property '''
        self.checkWorkspace()
        try:
            self.config.getWorkspaceById(self.workspaceId).setPropertyValue(prop, value)
        except WorkspaceConfigError, ex:
            raise GenericError(str(ex))

    def cmd_advanced_workspace_clean_cache(self):
        ''' Clean workspace bundle cache '''
        self.apt("clean")

    def cmd_workspace_list_changes(self, *bundle):
        ''' List changes to a workspace or specific bundle(s) '''
        self.checkWorkspace()
        if bundle and self.cmd_options.new:
            self.cmd_workspace_list_changes._optionparser.error("Option new and bundle argument are mutually exclusive")
        if self.cmd_options.bundlediff and self.cmd_options.new:
            self.cmd_workspace_list_changes._optionparser.error("Options new and bundlediff are mutually exclusive")

        wsPath = self.config.getWorkspaceById(self.workspaceId).path
        wsMetaPath = self.config.getWorkspaceMetadataPath(self.workspaceId)
        wsdiff = WorkspaceDiff(wsPath, wsMetaPath)
        optiondict = self.cmd_options.__dict__
        # If no mode options given, revert to defaults
        if (not self.cmd_options.new and
            not self.cmd_options.modified and
            not self.cmd_options.removed and
            not self.cmd_options.unmodified):
            optiondict["new"] = False
            optiondict["modified"] = True
            optiondict["removed"] = True
            optiondict["unmodified"] = False

        wsdiff.start(bundle, **optiondict)

    cmd_workspace_list_changes._optionparser = CommandOptionParser()
    cmd_workspace_list_changes._optionparser.add_option("-n", "--new", action="store_true",
                                                       help="List new files [default: False]")
    cmd_workspace_list_changes._optionparser.add_option("-m", "--modified", action="store_true",
                                                       help="List modified files [default: True]")
    cmd_workspace_list_changes._optionparser.add_option("-r", "--removed", action="store_true",
                                                       help="List removed files [default: True]")
    cmd_workspace_list_changes._optionparser.add_option("-u", "--unmodified", action="store_true",
                                                       help="List unmodified files [default: False]")
    cmd_workspace_list_changes._optionparser.add_option("-d", "--dir", action="append", dest="dirs", metavar="DIR",
                                                       help="Add dir which is used as a inclusion filter (you can set many)")
    cmd_workspace_list_changes._optionparser.add_option("-b", "--bundlediff", action="store_true",
                                                       help="List only bundles that have changed [default: %default]")
    cmd_workspace_list_changes._optionparser.set_defaults(new=False, modified=False, removed=False, unmodified=False, bundlediff=False)

    def getWorkspaceProperty(self, name):
        ''' Takes into account also global configuration '''
        workspace = self.config.getWorkspaceById(self.workspaceId)
        value = workspace.getPropertyValue(name, getDefault=False)
        if value is None:
            value = self.config.getPropertyValue(name)
        return value

    def cleanCache(self):
        if not self.getWorkspaceProperty("cache-bundles"):
            self.apt("clean")

    def getSourcesListPath(self):
        wsMetaPath = self.config.getWorkspaceMetadataPath(self.workspaceId)
        return os.path.join(wsMetaPath, "apt", "sources.list")