#
# Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
# All rights reserved.
# This component and the accompanying materials are made available
# under the terms of "Eclipse Public License v1.0"
# which accompanies this distribution, and is available
# at the URL "http://www.eclipse.org/legal/epl-v10.html".
#
# Initial Contributors:
# Nokia Corporation - initial contribution.
#
# Contributors:
#
# Description:
# Blocks style command line application class
#

from optparse import OptionParser, SUPPRESS_HELP
import logging
import pprint
import sys

import utils
from cmdlinecommand import CmdLineCommand
from generalexceptions import InvalidUserInput, InvalidFileInput, ExternalProcessError, GenericError
from UserConfiguration.UserConfiguration import BadWorkspace

class CmdlineApp(object):
    ''' Command line application '''

    def __init__(self, cmdlineCommand, prog, versionInfo):
        ''' cmdlineCommand is CmdLineCommand class '''
        self.cmdlineCommand = cmdlineCommand
        self.prog = prog
        self.versionInfo = versionInfo
        self.parser = None

    def initParser(self):
        self.parser = OptionParser(prog=self.prog, add_help_option=False)
        self.parser.disable_interspersed_args()
        self.parser.version = utils.getVersion(self.versionInfo)
        self.parser.add_option("--version", action="version", help="Show version information and exit")
        self.parser.add_option("--help", "-h", action="help", help="Show list of all commands and exit")
        self.parser.add_option("--verbose", "-v", action="count", help="Be verbose (add more for debug output)")
        self.parser.add_option("--vmax", action="store_true", help=SUPPRESS_HELP)
        self.parser.add_option("--psyco", action="store_true", help=SUPPRESS_HELP)
        self.parser.add_option("--quiet", "-q", action="count", help="Be quiet (use -qq for very quiet operation)")
        self.parser.set_defaults(verbose=0, quiet=0)

    def parseCmdline(self):
        self.initParser()
        # Show advanced help if help requested
        # If parse_args does not exit, help was not requested
        self.parser.set_usage(self.cmdlineCommand.getUsage(True))
        (options, args) = self.parser.parse_args()
        self.parser.set_usage(self.cmdlineCommand.getUsage(False))
        options.verbose = 100 if options.vmax else options.verbose
        return (options, args)

    def init(self):
        '''Override this for initialization.

        Logging is initialized before this method
        '''
        pass

    def run(self):
        options, args = self.parseCmdline()
        options.verbose = options.verbose - options.quiet
        try:
            utils.setupLogging(self.parser.get_version(), options.verbose, self.prog + "_log.txt")
        except GenericError, ex:
            sys.exit(str(ex))
        logging.debug("Command-line args: %s", " ".join(sys.argv[1:]))
        logging.debug("Options:\n %s", pprint.pformat(options.__dict__))

        if options.psyco:
            try:
                import psyco
                psyco.full()
            except ImportError:
                logging.warning("Psyco extension not available. Not using psyco optimization.")

        self.init()

        if len(args) == 0:
            self.parser.print_help()
        else:
            try:
                command = self.cmdlineCommand(options)
                command(args)
            except ExternalProcessError, ex:
                # If exception has output with it, external process did not
                # print any information to stdout so we have to do it here
                utils.error("Error: %s" % ex, bool(ex.output), ex.blocksRetcode)
            except CmdLineCommand.CommandNotAvailableError, ex:
                self.parser.print_help()
                print
                utils.error("Error: %s" % ex, retCode=2)
            except (CmdLineCommand.ArgumentError, InvalidUserInput), ex:
                print self.cmdlineCommand.getCommandUsage(args[0])
                print
                utils.error("Error: %s" % ex, retCode=2)
            except BadWorkspace, ex:
                utils.error("Workspace error: %s" % ex)
            except IOError, ex:
                utils.error("I/O Error: %s" % ex)
            except InvalidFileInput, ex:
                utils.error("Input file error: %s" % ex)
            except GenericError, ex:
                utils.error("Error: %s" % ex)
            except KeyboardInterrupt, ex: # pragma: no cover
                utils.error("User interrupt. Exiting...")
            except AssertionError, ex: # pragma: no cover
                utils.error("Assertion error: %s" % ex, True)
            except Exception, ex: # pragma: no cover
                utils.error("Unexpected error: %s" % ex, True)