#
# Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
# All rights reserved.
# This component and the accompanying materials are made available
# under the terms of "Eclipse Public License v1.0"
# which accompanies this distribution, and is available
# at the URL "http://www.eclipse.org/legal/epl-v10.html".
#
# Initial Contributors:
# Nokia Corporation - initial contribution.
#
# Contributors:
#
# Description:
# Multiprocess packager
#

'''
Packager reads BuildData to create PackageModel.Components;
these are passed to a PackageWriter to create actual packages.

Use multiprocessor module for concurrent packaging.
'''

import multiprocessing
import time
import itertools

from Blocks.Packaging.PackageWriter import DebWriter, ZipWriter, RpmWriter
from Blocks.Packaging.ComponentBuilder import ComponentBuilder, BuilderException
from Blocks.Packaging.Logging import Logging
from Blocks.Packaging.FileMapping import FileMap

class BuilderWorker(object):
    def __init__(self, builder, storage):
        self.builder = builder
        self.storage = storage

    def __call__(self, buildData):
        try:
            return self.builder.createComponent(self.storage, buildData)
        except KeyboardInterrupt, e:
            return e
        except BuilderException:
            raise
        except Exception, e:
            Logging.getLogger("pfw.packager.multiproc.builder").exception(e)
            raise

class WriterWorker(object):

    def __init__(self, writer, writerOptions, packageOutputDirectory,
                  storage, filemap, dependencyProcessors=None):
        self.writer = writer
        self.writerOptions = writerOptions
        self.packageOutputDirectory = packageOutputDirectory
        self.storage = storage
        self.globalFileMap = filemap
        self.dependencyProcessors = dependencyProcessors or []
        self.logger = None
        self.statusLogger = None

    def getWriter(self):
        if self.writer == "zip":
            return ZipWriter(self.writerOptions)
        elif self.writer == "rpm":
            return RpmWriter(self.writerOptions)
        elif self.writer == "deb":
            return DebWriter(self.writerOptions)
        else:
            return None

    def __call__(self, bdCompTuple):
        try:
            buildData, component = bdCompTuple
            self.logger = Logging.getLogger("pfw.packager.multiproc.writer")
            self.statusLogger = Logging.getStatusLogger()
            compName = buildData.getComponentName()
            self.statusLogger.info("Processing dependencies: %s" % compName)
            tick = time.time()
            for processorInitData in self.dependencyProcessors:
                processor = processorInitData[0]
                if len(processorInitData) == 1:
                    p = processor()
                else:
                    args = processorInitData[1]
                    p = processor(args)
                self.logger.debug("Running processor %s on %s" % (processor.__name__, compName))
                p(component, buildData, self.globalFileMap)
            tock = time.time()
            self.logger.info("Resolved dependencies for %s in %s seconds" % (compName, tock - tick))
            w = self.getWriter()
            if w:
                self.statusLogger.info("Sending to writer: %s" % compName)
                tick = time.time()
                written = w.write(component, buildData.getTargetRoot(), self.packageOutputDirectory, buildData.getSourceRoot())
                tock = time.time()
                self.logger.info("%s: wrote %s bundles to %s in %.3f seconds", compName, len(written), self.packageOutputDirectory, tock - tick)
            self.storage.writeMetaData(component)
            self.storage.writeFileMap(component)
            self.logger.debug("%s: writer done." % compName)
            return component
        except KeyboardInterrupt, e:
            return e
        except Exception, e:
            self.logger.exception(e)
            raise

class Packager(object):
    ''' Packager '''
    def __init__(self, storage, packageOutputDirectory,
                 maxWorkers=None, writer="deb", targetRules=None,
                 sourceRules=None, directives=None, writerOptions=None,
                 loadAllPreviousMetaData=True, keepGoing=True):
        '''
        @param storage: Storage info
        @type storage: Blocks.Packaging.Storage.PackagerStorage
        @param packageOutputDirectory: Where to write package files
        @type packageOutputDirectory: String
        @param maxWorkers: Maximum number of concurrent workers to launch. Default is number of CPUs.
        @type maxWorkers: None or Integer
        @param targetRules: Alternatives packaging rules for target files
        @type targetRules: String or file-like object
        @param sourceRules: Alternative packaging rules for source files
        @type sourceRules: String or file-like object
        @param directives: Alternative packaging directives
        @type directives: String or file-like object
        @param keepGoing: Ignore errors as best as possible
        '''
        self.logger = Logging.getLogger("pfw.packager.multiproc")
        self.globalFileMap = FileMap()
        self.nonBuiltComponentNames = []    # only filemapping is loaded for these
        self.builtComponentNames = []       # components built in this session
        self.storage = storage
        self.packageOutputDirectory = packageOutputDirectory
        self.maxWorkers = maxWorkers
        self.writer = writer
        self.writerOptions = writerOptions
        self.dependencyProcessors = []
        self.loadAllPreviousMetaData = loadAllPreviousMetaData # query storage for all components
        self.keepGoing = keepGoing
        self.targetRulesPath = targetRules
        self.sourceRulesPath = sourceRules
        self.directivesPath = directives
        self.times = [time.time()]
        self.rawBuildData = {}  # buildData indexed by component name
        self.components = [] # created components
        self.written = None  # written components
        self.builderWorker = None
        logSetup = (Logging.logLevel, Logging.logDir, Logging.logFormat, Logging.logTimeFormat,
                    Logging.doStatusLogs, None, Logging.lock)
        self.workerPool = multiprocessing.Pool(maxWorkers, Logging.setupLogHandling, logSetup)
        self.logger.info("Packager starting build #%s.", self.storage.getBuildId())
        self.logger.info("Packager storage: %s.", self.storage)

    def addProcessor(self, processorClass, arg=None):
        '''
        Register a dependency processor. All processors must be registered
        before starting writer workers.

        @param processorClass: The dependency processor class
        @type processorClass: Class
        @param arg: The init arg passed when instantiating procClass
        @type arg: Any
        '''
        self.dependencyProcessors.append((processorClass, arg))

    def addComponent(self, buildData):
        '''
        Add a component for packaging.

        @param buildData: A BuildData object
        @type buildData: BuildData
        '''
        if not self.builderWorker:
            builder = ComponentBuilder(self.targetRulesPath, self.sourceRulesPath, self.directivesPath, self.keepGoing)
            self.builderWorker = BuilderWorker(builder, self.storage)
        self.rawBuildData[buildData.getComponentName()] = buildData
        self.components.append(self.workerPool.apply_async(self.builderWorker, [buildData]))

    def addNonBuiltComponent(self, componentName):
        '''
        Use the filemap of a previously created component. This is normally
        automatic for all components found via the PackagerStorage instance.
        Only call this method explicitly if loadAllPreviousMetadata is set to
        False.

        @param componentName: Component name
        @type componentName: String
        '''
        self.nonBuiltComponentNames.append(componentName)

    def mapNonBuiltComponents(self):
        '''
        After builders have finished (and before writing), any non-built
        components that were not among the "normal" ones are added to the
        filemap. Normally the storage is queried and all the latest metadata
        is loaded - disable it by setting loadAllPreviousMetadata to False.
        '''
        if self.loadAllPreviousMetaData:
            for componentName in self.storage.getComponentNames():
                if componentName not in self.nonBuiltComponentNames:
                    self.nonBuiltComponentNames.append(componentName)

        ignoreComponents = set(itertools.chain.from_iterable(bd.previousNames for bd in self.rawBuildData.itervalues()))
        ignoreComponents.update(self.builtComponentNames)
        for componentName in self.nonBuiltComponentNames:
            if componentName not in ignoreComponents:
                self.logger.info("Mapping non-built component: %s" % componentName)
                oldCompLookupFile = self.storage.getLastFileMapFile(componentName)
                self.globalFileMap.loadFragment(oldCompLookupFile)

    def wait(self):
        '''
        Do everything needed after builders have been started, dependency
        processors have been registered and all components added.
        '''
        try:
            self.logger.debug("waiting for builder workers")
            while False in [r.ready() for r in self.components]:
                for result in self.components:
                    result.wait()
                    if result.successful():
                        self.logger.debug("success %s " % result.get().getName())
                        if result.get().getName() not in self.builtComponentNames:
                            self.globalFileMap.addAny(result.get())
                            self.builtComponentNames.append(result.get().getName())
                    else:
                        # raise the exception
                        result.get()
            self.logger.debug("built all components")
            self.times.append(time.time())
            self.logger.info('Deps processing: %.3fs' % (self.times[1]-self.times[0]))
            self.mapNonBuiltComponents()
            self.times.append(time.time())
            self.logger.info('Map non built components: %.3fs' % (self.times[2]-self.times[1]))
            if not self.builtComponentNames:
                self.logger.warning("exiting without starting writer workers - no components were built.")
            else:
                writerWorker = WriterWorker(self.writer, self.writerOptions, self.packageOutputDirectory,
                                       self.storage, self.globalFileMap, self.dependencyProcessors)
                bdCompList = [(self.rawBuildData[result.get().getName()], result.get()) for result in self.components if result.successful()]
                self.written = self.workerPool.map_async(writerWorker, bdCompList)
                #for result in [result.get() for result in self.components if result.successful()]:
                #    writerWorker(self.rawBuildData[result.getName()], result)
                self.logger.debug("waiting for writer workers")
                self.written.wait()
                if not self.written.successful():
                    # raise exceptions
                    self.written.get()
                self.times.append(time.time())
                self.logger.info('Create bundles: %.3fs' % (self.times[3]-self.times[2]))

            self.logger.info('Execution summary:')
            self.logger.info('Deps processing: %.3fs' % (self.times[1]-self.times[0]))
            self.logger.info('Map non built components: %.3fs' % (self.times[2]-self.times[1]))
            if len(self.times) > 3:
                self.logger.info('Create bundles: %.3fs' % (self.times[3]-self.times[2]))
            self.logger.info("done")
        except KeyboardInterrupt:
            self.logger.critical("Terminated by keyboard interrupt")
            self.workerPool.terminate()
            self.workerPool.join()
        finally:
            self.workerPool.close()
            self.workerPool.join()